/*
 * This is the ARM Android incarnation of arch-dependent OS-dependent
 * routines. See also "android-os.c". Most of the OS-dependent things
 * are implemented in arm-linux-os.c
 */

/*
 * This software is part of the SBCL system. See the README file for
 * more information.
 *
 * This software is derived from the CMU CL system, which was
 * written at Carnegie Mellon University and released into the
 * public domain. The software is in the public domain and is
 * provided with absolutely no warranty. See the COPYING and CREDITS
 * files for more information.
 */
