dynbind.o dynbind.d : dynbind.c genesis/sbcl.h runtime.h lispobj.h align.h globals.h \
 os.h target-os.h target-arch-os.h target-arch.h thread.h \
 genesis/symbol.h lispobj.h symbol-tls.inc align.h genesis/vector.h \
 genesis/static-symbols.h genesis/thread.h gencgc-alloc-region.h \
 genesis/vector.h interrupt.h validate.h gc-typedefs.h interr.h
