/*
 * This is a machine-generated file. Please do not edit it by hand.
 *
 * This file contains low-level information about the
 * internals of a particular version and configuration
 * of SBCL. It is used by the C compiler to create a runtime
 * support environment, an executable program in the host
 * operating system's native format, which can then be used to
 * load and run 'core' files, which are basically programs
 * in SBCL's own format.
 */
#ifndef SBCL_GENESIS_NUMBER_TYPES
#define SBCL_GENESIS_NUMBER_TYPES
#include "lispobj.h"
struct bignum {
    lispobj header;
    sword_t digits[1];
};
struct ratio {
    lispobj header;
    lispobj numerator;
    lispobj denominator;
};
struct double_float {
    lispobj header;
    double value;
};
struct complex {
    lispobj header;
    lispobj real;
    lispobj imag;
};
struct complex_single_float {
    lispobj header;
    struct { float data[2]; }  data;
};
struct complex_double_float {
    lispobj header;
    lispobj filler;
    double real;
    double imag;
};
struct simd_pack {
    lispobj header;
    lispobj tag;
    long lo_value;
    long hi_value;
};
struct simd_pack_256 {
    lispobj header;
    lispobj tag;
    long p0;
    long p1;
    long p2;
    long p3;
};
struct simd_pack_512 {
    lispobj header;
    lispobj tag;
    long p0;
    long p1;
    long p2;
    long p3;
    long p4;
    long p5;
    long p6;
    long p7;
};
struct simd_pack_512_mask {
    lispobj header;
    long value;
};
#endif
