alloc.o alloc.d : alloc.c gc.h genesis/sbcl.h gc-assert.h interr.h os.h runtime.h \
 lispobj.h align.h target-os.h target-arch-os.h target-arch.h \
 gc-typedefs.h globals.h immobile-space.h gencgc-impl.h \
 gencgc-alloc-region.h genesis/cardmarks.h lispobj.h genesis/cons.h \
 genesis/weak-pointer.h hopscotch.h code.h genesis/closure.h \
 forwarding-ptr.inc pseudo-atomic.h genesis/thread.h \
 gencgc-alloc-region.h genesis/static-symbols.h genesis/symbol.h \
 symbol-tls.inc align.h genesis/vector.h genesis/gc-tables.h \
 genesis/vector.h arch.h thread.h interrupt.h validate.h \
 genesis/thread-init.inc
