parse.o parse.d : parse.c genesis/sbcl.h runtime.h lispobj.h align.h globals.h \
 os.h target-os.h target-arch-os.h target-arch.h vars.h parse.h \
 interrupt.h genesis/static-symbols.h lispobj.h thread.h genesis/symbol.h \
 symbol-tls.inc align.h genesis/vector.h genesis/thread.h \
 gencgc-alloc-region.h genesis/vector.h validate.h gc-typedefs.h \
 lispregs.h target-lispregs.h genesis/regnames.h arch.h search.h \
 genesis/closure.h
