x86-64-linux-os.o x86-64-linux-os.d : x86-64-linux-os.c os.h genesis/sbcl.h runtime.h \
 lispobj.h align.h target-os.h target-arch-os.h target-arch.h arch.h \
 thread.h globals.h genesis/symbol.h lispobj.h symbol-tls.inc align.h \
 genesis/vector.h genesis/static-symbols.h genesis/thread.h \
 gencgc-alloc-region.h genesis/vector.h interrupt.h validate.h \
 gc-typedefs.h interr.h lispregs.h target-lispregs.h genesis/regnames.h
