(in-package :sb-manual)

(defsection @contributed-modules (:title "Contributed Modules")
  "SBCL comes with a number of modules that are not part of the core
  system. These are loaded via `(REQUIRE :<MODULENAME>)`
  (see @CUSTOMIZATION-HOOKS-FOR-USERS). This section contains
  documentation (or pointers to documentation) for some of the
  contributed modules."
  (@sb-aclrepl section)
  (@sb-concurrency section)
  (@sb-cover section)
  (@sb-grovel section)
  (@sb-introspect section)
  (@sb-manual section)
  (@sb-md5 section)
  (@sb-posix section)
  (@sb-queue section)
  (@sb-rotate-byte section)
  (@sb-simd section))

